/*
A GLUT program using OpenGL, GLUI, and GLSL.

Orion Sky Lawlor, olawlor@acm.org, 2007/01/18 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;

/* Dr. Lawlor's GLSL utilities */
#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"


/* Variables updated by the GUI */
float distance=2.0; /* origin-to-camera distance */
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 modelMat(1.0), viewMat(1.0);

/* Draw a big green ground polygon */
void draw_ground(void) {
	double ground=100.0; /* extent of ground in X and Z */
	double groundY=0.0;
	glColor4f(0.1,0.4,0.1,1); /* green */
	glNormal3f(0,1,0); /* facing straight up */
	oglTextureQuad(
		oglVector3d(-ground,groundY,-ground), oglVector3d(+ground,groundY,-ground), 
		oglVector3d(-ground,groundY,+ground), oglVector3d(+ground,groundY,+ground)
	);
}

/* Draw coordinate axes--red, green, blue squished-spheres of unit length */
void draw_axes(void) {
	for (int axis=0;axis<3;axis++) {
		glPushMatrix();
		
		vec3 shift=vec3(0.0); shift[axis]=0.5;
		glTranslatef(shift.x,shift.y,shift.z);
		
		vec3 size=vec3(0.1);   size[axis]=0.5;
		glScalef(size.x,size.y,size.z);
		
		vec3 color=vec3(0.2);   color[axis]=0.8;
		glColor3fv(color);
		
		glutSolidSphere(1.0,20,20);
		glPopMatrix();
	}
}

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	// Clear the color and depth buffers
	glClearColor(0.2,0.4,0.8,0.0); /* sky blue */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); /* actually clear color and depth */
	
	/* Depth testing */
	glEnable(GL_DEPTH_TEST);
	
	/* Alpha blending */
	glEnable(GL_BLEND);
	           /*  src color  + dest color  */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	// Load all needed matrices into OpenGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /*<- clean out any old leftover matrices */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		win_w/(float)win_h, /* viewport's aspect ratio: always width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	glTranslatef(0,0,-distance); /* push world origin down the *unrotated* Z axis */
	glMultMatrixf(&viewMat[0][0]); /* camera orientation: could also go under GL_MODELVIEW */
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	/* Set up programmable shader. */
	static oglProgramObject *p=oglProgramFromFiles("vertex.txt","fragment.txt");
	p->begin();
	p->set("colorscale",1.0f); /* normal case-- don't mess with colors */
	
	/* Draw a big green ground polygon */
	draw_ground();
	
	// Draw the world 
	glMultMatrixf(&modelMat[0][0]);
	glColor4f(1,1,0,1); /* yellow */
	glutSolidTorus(0.2,0.5,40,60);
	
	// Draw current coordinate axes
	draw_axes();
	
	p->end();
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Lighting Demo");

	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	// Make some sliders (this is the only GLUI code you'll need!)
	GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
		GLUI_SUBWINDOW_TOP);
	/* This causes the main window to be redrawn when a GLUI control changes */
	glui->set_main_gfx_window( main_window );

	GLUI_Spinner *spin_zoom = new GLUI_Spinner( glui, "Distance:", &distance);
	spin_zoom->set_float_limits(0.0, 10.0);
	
	/* Set up view matrix so it's tilted forward a bit */
	vec3 up=normalize(vec3(0,1,0.2));
	viewMat[1]=vec4(up,0.0);
	viewMat[2]=vec4(vec3(up.x,-up.z,up.y),0.0);
	
	new GLUI_Column(glui);
	new GLUI_Rotation(glui,"Model:",&modelMat[0][0]);
	new GLUI_Column(glui,false);
	new GLUI_Rotation(glui,"View:",&viewMat[0][0]);
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
